;;;(load "initogl")
(pack (oglwin '.gl :width 300 :height 300) 
    :side 'left :expand 'yes :fill 'both)


; Set up default gl values
( .gl 'eval 
	:matrixmode 'projection 
	:loadidentity 
	:ortho -5 5 -5 5 -5 5
	:material 'frontandback 'ambientanddiffuse 0.7 0.7 0.7 
	:material 'frontandback 'specular 1 1 1 
	:material 'frontandback 'shininess 100 
	:enable 'autonormal 
	:shademodel 'smooth 
	:enable 'light0 
	:enable 'lighting 
	:enable 'normalize 
	:enable 'depthtest)

;
; Set up default nurbs values
;

; Init nurbs surface data
(define  cpoints ())
(do ((u 0 (+ u 1)))
     ((>= u  4))
    (do (( v 0(+ v 1))) 
        ((>= v 4))
	(set! cpoints (append cpoints 
           (list (* (- u 1.5)2.0) (* (- v 1.5) 2.0)))) 
	(if (and (or (= u 1)(= u 2)) (or (= v 1)(= v 2)))
	    (set! cpoints (append cpoints '(3)))
	    (set! cpoints (append cpoints '(-3)))
        )	
    )
)

(define knottype 'uniform)
(define order 4)
(define culling 'f)
(define sampling 50)
(define  displaymode 'fill)

; Create control point mesh display
(define (cpoint  i j )
    ; returns control point (i,j) 
  (let ((k  (* (+ (* i 4) j) 3)))
       (lrange cpoints k (+ k 2))
  )
)

(define l ())
(do ((i 0 (1+  i)))
    ((>= i  4))
    (set!  l (append l '(:begin 'linestrip))) 
    (do ((j 0 (1+ j)))
        ((>= j  4))
	(set! l (append l `(:vertex ,@(cpoint i j))))
    )
   (set! l (append l '(:end)))
)

(do ((i 0(1+ i)))
    ((>= i  4))
    (set! l (append l '(:begin 'linestrip))) 
    (do ((j 0 (1+ j)))
        ((>=  j  4))
	(set! l (append l `(:vertex ,@(cpoint j i))))
    )
   (set! l (append l '(:end)))
)

(define  cpointmesh (eval `(.gl 'newlist ,@l)()))
(define knottype 'uniform)
(define order 4)
(define culling 'f)
(define sampling 50)
(define  displaymode 'fill)

;(define  cpointdisplay (.gl 'newlist :call cpointmesh))

(define  cpointdisplay (.gl 'newlist :call cpointmesh))
(define  cpointflag #t)

; Compute the nurbs picture
(define knots ())
(define nurbs ())

(define (computenurbs )
    ;global cpoints sampling culling displaymode nurbs order knottype cpointdisplay
  (let ((n 0)(k 0)(j 0)(d 0))
    (if (number? nurbs)
        (catch (.gl 'deletelist nurbs))
    )
    (set!  knots ())
    (set!  n (+ order 4))
    (if (equal? knottype 'uniform)
      (begin
	(set!  k (/  1.0 (- n 1)))
	(do ((i 0 (1+ i)))
            ((>= i n))
	    (set! knots (append knots (list (* k i))))
	)
      )
    ; else 
     (begin
	(set!  j (- n (* order 2)))
	(set!  d (/ 1.0 (1+ j)))
	(do ((i 0 (1+ i)))
            ((>= i j))
	    (set! knots (append knots (list (* (1+ i) d))))
	)
	(do (( i 0(1+ i)))
            ((>= i order))
	    (set! knots (append knots '(1)))
	    (set! knots (cons 0  knots ))
	)
    )
  )
   ( set!  nurbs (eval `(.gl 'nurbssurface 
	    :sampling sampling :culling culling :uorder order 
	    :vorder order :displaymode displaymode
	    :vknots ,@knots :uknots ,@knots :controlpoints ,@cpoints)() ))
   ( .gl 'main :clear 'colorbuffer 'depthbuffer 
	    :enable 'lighting :linewidth 1 :call nurbs
	    :disable 'lighting :color 1 1 0 
            :linewidth 2 :call cpointdisplay)
))
;;;  (computenurbs)
   
; Set up rotation bindings
(define  xrot 0)
(define  yrot 0)
(define  mouse-x 0)
(define  mouse-y 0)
(define  pmode 'fill)

(bind .gl "<Button-1>" (lambda (x y)( set! mouse-x x)(set! mouse-y y)))
(bind .gl "<B1-Motion>" (lambda (x y) 
    (set! xrot(+  xrot (- x mouse-x))) 
    (set! yrot(+  yrot (- y mouse-y))) 
    (set! mouse-x x)
    (set! mouse-y y)
    (.gl 'eval :matrixmode 'modelview 
            :loadidentity 
            :rotate yrot 1 0 0
            :rotate xrot 0 1 0)
    (.gl 'redraw)
))

;
; Set up rest of interface
;
(pack (frame '.cmd) :side 'right :fill 'y)

; Set up Command buttons
(pack (frame '.cmd.button) :fill 'x)
(pack (button '.cmd.button.quit :text "Quit" :command "exit")
         :fill 'x :side 'left)
(pack (button '.cmd.button.recompute :text "Recompute"
        :command '(computenurbs) )
	:fill 'x :side 'left)
(pack (checkbutton '.cmd.button.showframe :text "Frame" :var 'cpointflag
     :command '(begin
      (if  cpointflag 
        (.gl 'newlist cpointdisplay :call cpointmesh)
      ; else {
        (.gl 'newlist cpointdisplay)
      )
    (.gl 'redraw)
   )  )
  :fill 'x :side 'left)
        
; Set up display mode interface
(pack (frame '.cmd.dmode) :fill 'both)
(pack (label '.cmd.dmode.label :text "Display Mode"):side 'left)
(pack (frame '.cmd.dmode.value :bd 2 :relief 'sunken) :side 'right)
(foreach dmode '(fill outlinepolygon outlinepatch)
    ;;;(format #t "debug: ~s ~%" dmode)
    (pack (radiobutton (tk-name '.cmd.dmode.value. dmode)
             :text dmode :var 'displaymode 
        :value dmode :anchor 'w 
        :command '(begin (computenurbs)(.gl 'redraw)))
        :fill 'x)
)

; Set up sampling tolerance interface
(pack (frame '.cmd.stol) :fill 'both)
(pack (label '.cmd.stol.label :text "Sampling Tolerance") :side 'left)
(pack (frame '.cmd.stol.value :bd 2 :relief 'sunken) :side 'right)
(foreach stol '(10 25 50)
    (pack (radiobutton (tk-name '.cmd.stol.value. stol)
              :text stol :var 'sampling
	:value stol :anchor 'w :command '(begin (computenurbs)(.gl 'redraw)))
	  :fill 'x)
)

; Set up knot type interface
(pack (frame '.cmd.ktype) :fill 'both)
(pack (label '.cmd.ktype.label :text "Knot Distribution") :side 'left)
(pack (frame '.cmd.ktype.value :bd 2 :relief 'sunken) :side 'right)
(foreach ktyp '(uniform non-uniform)
   ( pack (radiobutton (tk-name '.cmd.ktype.value. ktyp)
            :text ktyp :var 'knottype
      :value ktyp :anchor 'w :command '(begin (computenurbs)(.gl 'redraw)))
	    :fill 'x)
)

; Set up Order interface
(pack (frame '.cmd.order) :fill 'both)
(pack (label '.cmd.order.label :text "Polynomial Order") :side 'left)
(pack (frame '.cmd.order.value :bd 2 :relief 'sunken) :side 'right)
(foreach ord '(2 3 4)
   (pack (radiobutton (tk-name '.cmd.order.value. ord)
            :text ord :var 'order
      :value ord :anchor 'w :command '(begin (computenurbs)(.gl 'redraw)))
	    :fill 'x)
)

(computenurbs) 


