/*** floor.c
***/
#define  FLOOR_C

#include "lisp.h"
#include <math.h>

#define minusp(x)  gsign(x)
extern double get_float();
extern ptr _float(),float_div();
extern mpz_t mpzwork1,mpzwork2;
extern ptr gensub(),genadd(),genmul(),gendiv();

ptr negative();

/***   float -> floor x ***/
ptr ffloor(ptr x)
 {
    double s;
    /** printf("ffloor %f\n",FLONUM(x)); ***/
    s=floor(FLONUM(x));
     /** return MAKEDBL(s); ***/
     /***  04.1.20  ***/
   return d2rational(s);    /*** return rational  ***/
 }
 
/***  rational-> floor ***/

ptr rfloor(y)
  Reg ptr y;
{
  if(INTEGERP(y)) return y;
  /*** ratio  **/
  psh(y);
  mpz_fdiv_q(mpzwork1,mpq_numref(RATIONAL(y)),mpq_denref(RATIONAL(y)));
  y= mpz_check16(mpzwork1);
  pull();
  return y;
}

ptr rtruncate(y)
  Reg ptr y;
{
  if(INTEGERP(y)) return y;
  psh(y);
  mpz_tdiv_q(mpzwork1,mpq_numref(RATIONAL(y)),mpq_denref(RATIONAL(y)));
  y= mpz_check16(mpzwork1);
  pull();
  return y;
}

ptr _ftruncate(x)
  Reg ptr x;
{
  Reg double s;
        if((s=FLONUM(x))<0)
          {
            s=floor(-s);
          return d2rational(-s);    /*** return rational  ***/
            /** return MAKEDBL(-s);  ***/
          }
       return ffloor(x);       
}

ptr rceiling(y)
  Reg ptr y;
{
  if(INTEGERP(y)) return y;
  psh(y);
  mpz_cdiv_q(mpzwork1,mpq_numref(RATIONAL(y)),mpq_denref(RATIONAL(y)));
  y= mpz_check16(mpzwork1);
  return y;
}

ptr fceiling(x)
  Reg ptr x;
{
  Reg double s;
        s=FLONUM(x);
        s=floor(-s);
          return d2rational(-s);    /*** return rational  ***/
    /***  return MAKEDBL(-s);  ***/
}

ptr rround(x)
  Reg ptr x;
{
  Reg ptr *s=stack;
  
  if(INTEGERP(x)) return x;
  *stack++=gendiv(ONE,TWO);
  x=*s=genadd(*s,x);
   x=rfloor(x);
     pull();
     return x;
}

ptr fround(x)
  Reg ptr x;
{
  Reg double s;
        s=FLONUM(x);
        s=floor(s+0.5);
          return d2rational(s);    /*** return rational  ***/
     /*** return MAKEDBL(s);  ***/
}

ptr genfix(narg,fix_ratio,fix_float)
    /*  general form of floor,.... */
   int narg;
    ptr (*fix_ratio)();    /** function of fix(rational) ***/
    ptr (*fix_float)();    /** function of fix(float) ***/
 {
    Reg ptr x,y,*s;
    s=stack-narg;
    x=*s++;
    narg--;
    if(narg==0) y=ONE;
    else{
      y=*s++;
      checknarg(narg-1);
    }
    s=stack;
    stack+=2;
    if(INTEGERP(x)||RATIOP(x))
     {
       if(y==ONE)
        {
          *s=(*fix_ratio)(x);
           y=gensub(x,*s);      
        }
       else if(INTEGERP(y)||RATIOP(y))
        {
          *s=gendiv(x,y);
          *s=(*fix_ratio)(*s);
          s[1]=genmul(*s,y);
          y=gensub(x,s[1]);
        }
       else if(FLOATP(y))
        {
          if(get_float(y)==1.0)
           {
            *s=(*fix_ratio)(x);
             *s=_float(*s);
             y=gensub(x,*s);      
           }
         else
           {
            *s=gendiv(x,y);
            *s=(*fix_float)(*s);
            s[1]=genmul(*s,y);
            y=gensub(x,s[1]);
           }
        }
       else
          error(E_notnum,y);
    }   
   else if(FLOATP(x))
    {
       if(y==ONE)
        {
          *s=(*fix_float)(x);
           y=gensub(x,*s);      
        }
       else if(INTEGERP(y)||RATIOP(y))
        {
          y=s[1]=_float(y);
          *s=floatdiv(x,y);
          *s=(*fix_float)(*s);
          s[1]=genmul(*s,y);
          y=floatsub(x,s[1]);
        }
       else if(FLOATP(y))
        {
          if(get_float(y)==1.0)
           {
            *s=(*fix_float)(x);
            *s=_float(*s);
            y=gensub(x,*s);      
           }
         else
           {
            *s=gendiv(x,y);
            *s=(*fix_float)(*s);
            s[1]=genmul(*s,y);
            y=floatsub(x,s[1]);
           }
        }
       else
          error(E_notnum,y);
    }   
  else
     error(E_notnum,x);
    x=*s;
    sp_pull(2);
    
    /** return value2(x,y); **/
    return x;
  }

ptr s_floor(int narg,ptr env)
    /*   expr 2 & return two value 90.6.11 */
 {
   return genfix(narg,rfloor,ffloor);    
 }
ptr s_truncate(int narg,ptr env)
    
 {
  
   return genfix(narg,rtruncate,_ftruncate);    
 }
ptr s_round(int narg,ptr env)
    
 {
  
   return genfix(narg,rround,fround);    
 }

ptr s_ceiling(int narg,ptr env)
    
 {
  
   return genfix(narg,rceiling,fceiling);    
 }
       
ptr negative(p)  /*  hakai teki  */
  Reg ptr p;
 {
   if(INTP(p))
     {
        return (ptr)((~(long)p)+3);
     }
  else if(BIGINTP(p))
    {
      ATTR(p)=ATTR(p)^1;
    }
  else if(RATIOP(p))
    {
      CAR(p)=negative(CAR(p));
    }
  else if(FLOATP(p))
    {
      CAR(p)=(ptr)(((long)CAR(p))^0x80000000);  /** 32 bit **/
    }
  else
     error(E_notnum,p);
   return p;
 }     
/*** 91.11.25  ffloor  ***/
ptr genfixf(narg,fix_ratio,fix_float)
    /*  general form of floor,.... */
    int narg;
    ptr (*fix_ratio)();    /** function of fix(rational) ***/
    ptr (*fix_float)();    /** function of fix(float) ***/
 {
    Reg ptr x,y,*s;
    if(narg==1){
      x=*(stack-1);
       y=ONE;
    }
    else if (narg==2){
      x=*(stack-2);
      y=*(stack-1);
    }
    else  checknarg(narg-2);
           
    s=stack;
    stack+=2;
    *s=x=_float(x);
       if(y==ONE)
        {
          *s=(*fix_float)(x);
           y=gensub(x,*s);      
        }
       else if(INTEGERP(y)||RATIOP(y))
        {
          y=s[1]=_float(y);
          *s=floatdiv(x,y);
          *s=(*fix_float)(*s);
          s[1]=genmul(*s,y);
          y=floatsub(x,s[1]);
        }
       else if(FLOATP(y))
        {
          if(get_float(y)==1.0)
           {
            *s=(*fix_float)(x);
            *s=_float(*s);
            y=gensub(x,*s);      
           }
         else
           {
            *s=gendiv(x,y);
            *s=(*fix_float)(*s);
            s[1]=genmul(*s,y);
            y=floatsub(x,s[1]);
           }
        }
       else
          error(E_notnum,y);
    x=_float(*s);
    sp_pull(2);
   return x;
}

ptr s_ffloor(int narg,ptr env)
    /*   expr 2 & return two value 90.6.11 */
    
 {
  
   return genfixf(narg,rfloor,ffloor);    
 }
ptr s_ftruncate(int narg,ptr env)
    
 {
  
   return genfixf(narg,rtruncate,_ftruncate);    
 }
ptr s_fround(int narg,ptr env)
    
 {
  
   return genfixf(narg,rround,fround);    
 }

ptr s_fceiling(int narg,ptr env)
    
 {
  
   return genfixf(narg,rceiling,fceiling);    
 }
       
 


