/*
 *
 * set . c			-- Set procedures
 *
 * Copyright  2004  Sadao Sato
 * 
 *
 * Permission to use, copy, modify, distribute,and license this
 * software and its documentation for any purpose is hereby granted,
 * provided that existing copyright notices are retained in all
 * copies and that this notice is included verbatim in any
 * distributions.  No written agreement, license, or royalty fee is
 * required for any of the authorized uses.
 * This software is provided ``AS IS'' without express or implied
 * warranty.
 *
 *
 *           Author: Sadao Sato
 *    Creation date: 12-Dec-2002
 * Last file update: 09-June-2004
 */

#include "lisp.h"
#include "string.h"

extern ptr *ptr_alloc();
extern ptr equal(),CONS();

static ptr set_equal(),storage2list();
static int add_list_set();

static ptr *found;

static ptr equal_element(ptr x,ptr y)
{
  if(equal(x,y)==a_t) return a_t;
  if(SREGP(x)||SREGP(y)) return a_f;
  if(CONSP(x)&&CONSP(y)) return set_equal(x,y);
  return a_f;
}

int set_member(ptr x,ptr *q,int len)
 {
   while(len--){
     if(equal_element(x,*q++)==a_t){
           found=q-1;
           return YES;
      }
   }
   return NO;
 }

ptr s_set_remove(int narg,ptr env)
{
  Reg ptr x,set,*p;
  Reg int n;
  checknarg(narg-2);
   x=*(stack-2);
   set=*(stack-1);
   if(NULLP(set)) return set;
   if(!CONSP(set)) error(E_notlist,set);
   p=ptr_alloc(_length(set),0);
   n=add_list_set(set,p,0);
   if(set_member(x,p,n)){
      *found=a_unbdd;
      set=storage2list(p,n);
   }
    Myfree(p);
    return set;
}

ptr s_set_member(int narg,ptr env)
{
  Reg ptr x,set;
  checknarg(narg-2);
   x=*(stack-2);
   set=*(stack-1);
   if(NULLP(set)) return a_f;
   if(!CONSP(set)) error(E_notlist,set);
   while(CONSP(set)){
     if(equal_element(x,CAR(set))==a_t) return a_t;
     set=CDR(set);
   }
   return a_f;
}


/***   add_list_set
  q,len --- set
  &  strage must continue to q+len+length(lis)
   return total elements
 ***/
 
static int  add_list_set(ptr lis,ptr *q,int len)
{
   Reg ptr x,*r;
   r=q+len;
  loop:
   if(NULLP(lis)) return len;
   x=CAR(lis);
   lis=CDR(lis);
   if(set_member(x,q,len)) goto loop;
   *r++=x;
   len++;
   goto  loop;
}


/*** 
   dest storage also must have  len 
   set -> set-lis   (by undef  )
   dest ->  intersection
   return  size of intersection
***/
static int  diff_list_set(ptr lis,ptr *q,int len, ptr *dest)
{
   Reg ptr x,*r;
   Reg int n=0;
  loop:
   if(NULLP(lis)) return n;
   x=CAR(lis);
   lis=CDR(lis);
   if(!set_member(x,q,len)) goto loop;
   *dest++=x;
   n++;
   *found=a_unbdd;   
   goto  loop;
}


static ptr storage2list(ptr *p, int len)
{
  int j,first;
  ptr z, tmp,x;
  z = NIL;
  first=1;
  psh(z);
  for (j=0; j<len; j++) {
    x=*p++;
    if(x!=a_unbdd)
      if (first){
        tmp = z = CONS(x, NIL);
        first=0;
        }
       else
        tmp = CDR(tmp) = CONS(x, NIL);
  }
  pull();
  return z;
}

static void  rem_dupl(ptr *p, int n)
{
   ptr y,z;
   int i,j;
   ptr *q=p,*r=p;
  for(i=0;i<n-1;i++){
   y=*p++;
   if(y==a_unbdd) continue;
   q=p;
    for(j=1;j<n-i;j++){
     z=*q;
     if(z!=a_unbdd)
       if(equal_element(y,z)==a_t) *q=a_unbdd;
     q++;
   }
 }
}

static int  rem_reduce(ptr *p, int n)
{
   ptr y,z;
   int i,j;
   ptr *q=p;
   j=0;
  for(i=0;i<n;i++){
   y=*p++;
   if(y==a_unbdd) continue;
   *q++=y;
   j++;
   }
 return j;
}


ptr s_list2set(int narg,ptr env)
{
 ptr  x;
 int n;
 ptr *p;
 checknarg(narg-1);
 x=*(stack-1);
 if(x==NIL) return x;
 if(!CONSP(x))  error(E_notlist, x);
 n= _length(x);
 if(n==1) return x;
/* v= vector(x, _length(x)); */
 p=ptr_alloc(n,0);
 n=add_list_set(x,p,0);
   x=storage2list(p,n);
  Myfree(p);
   return x;
}

static ptr set_union(ptr *sets, int length)
{
  int len,n,m,i,j,storelen,mem;
  ptr  x,y,v,z;
  ptr *p,*q;
  if (length==0) return NIL;
  x=*sets++;
  if (length==1) return x;
  if (NULLP(x)) return set_union(sets,length-1);
  if (!CONSP(x)) Err("set-union: not a list",E_notlist, x);
  length--;
  mem=j=0;
  while(j<length){
    y=*sets++;
    if(NULLP(y)) {
        j++;
        if(j==length){
             if(mem==0) return x;
             break;
         }
        continue;
     }
    if (!CONSP(y)){
           if(j>0) Myfree(q);
           Err("set-union: not a list",E_notlist, y);
    }
    m=_length(y);
    if(mem==0){
       mem=len=(n=_length(x))+m;
        p=q=ptr_alloc(len,a_unbdd);
         for(i=0;i<n;i++){
         *p++=CAR(x);
          x=CDR(x);
         }
      storelen=len;
      }
    else
     {
      len=n+m;
      if(len>storelen){
         storelen=len;
         p=ptr_alloc(len,a_unbdd);
         memcpy(p, q, n * sizeof(ptr));
         Myfree(q);
         q=p;
       }
      p=q+n;
     }
     n=add_list_set(y,q,n);
      j++;
   }
    z= storage2list(q,n);
  Myfree(q);
  return z;
}

ptr s_set_union(int narg,ptr env){
  return set_union(stack-narg,narg);
}

static ptr set_intersection(ptr *sets, int len)
{
  int n;
  ptr x,y;
  ptr *p,*dest,*s;
    if(len==0)  return NIL;
   x=*sets++;
    if((len==1)||NULLP(x))  return x;
  
  if (!CONSP(x)) Err("set-intersection: not a list",E_notlist, x);
  n=_length(x);
   p=(ptr *) ptr_alloc(n,a_unbdd);
   dest=(ptr *) ptr_alloc(n,a_unbdd);
   n=add_list_set(x,p,0);
  len--;
  for(;len>0;len--){
   y=*sets++;
   if (NULLP(y)) goto Retnil;
   if (!CONSP(y)) goto Error; 
   n=diff_list_set(y,p,n,dest);
   if(n==0) goto Retnil;
  /*** buffer change **/
  s=p;
  p=dest;
  dest=s;
 }
   x=storage2list(p,n);
   Myfree(p);
   Myfree(dest);
   return x;
    
  Retnil:
    Myfree(p);
   Myfree(dest);
     return NIL;
  Error:
   Myfree(p);
   Myfree(dest);
     Err("set-intersection: not a list", y);
  return UNDEFINED; /* never reached */
} 

ptr s_set_intersection(int narg,ptr env){
  return set_intersection(stack-narg,narg);
}

static ptr set_difference(ptr *sets, int len)
{
  int n;
  ptr x,y;
  ptr *p,*dest;
  if(len==0) return NIL;
  x=*sets++;
  if(len==1) return x; 
  if (NULLP(x)) return x;
  if (!CONSP(x)) Err("set-differnce: not a list",E_notlist, x);
  n=_length(x);
  p=(ptr *) ptr_alloc(n,a_unbdd); /* VECT(v); */
  dest=(ptr *) ptr_alloc(n,a_unbdd); /* VECT(v); */
  n=add_list_set(x,p,0);
  len--;
  while(len--){
    y=*sets++;
    if(NULLP(y)) continue;
    if (!CONSP(y)) {
             Myfree(p);
             Myfree(dest);
            Err("set-differnece: not a list",E_notlist, y);
     }
    diff_list_set(y,p,n,dest);
    n=rem_reduce(p,n);
    if(n==0){
             Myfree(p);
             Myfree(dest);
         return NIL;
    }
 }
 x=storage2list(p,n);
 Myfree(p);
 Myfree(dest);
 return x;
}

ptr s_set_difference(int narg,ptr env){
  return set_difference(stack-narg,narg);
}

static ptr subset(ptr  x, ptr  y)
{
   /***  x is subset of y?  ***/
  int n,m,ans;
  ptr  v,z,tmp;
  ptr *p,*q;
  if (NULLP(x)) return Truth;
  if (NULLP(y)) return Ntruth;
  if (!CONSP(x)) Err("set-subset: not a list",E_notlist, x);
  if (!CONSP(y)) Err("set-subset: not a list",E_notlist, y);
  n=_length(x);
  m=_length(y);
  if(n>m) return Ntruth;
  p=(ptr *) ptr_alloc(m,a_unbdd);
  add_list_set(y,p,0);
  ans=YES;
  while(n--){
    if(set_member(CAR(x),p,m)!=YES){
         ans=NO;
         break;
     }
   x=CDR(x);
  }
    Myfree(p);
  return PRED(ans);
}

/****  x,y must list ***/
static ptr set_equal(ptr x, ptr  y)
{ /***  x is subset of y?  ***/
  int n,m,i,j,del;
  ptr  v,z,tmp;
  ptr *p,*q;
  if (NULLP(x)) {
         if(NULLP(y)) return Truth;
         return Ntruth;
    }
  if (NULLP(y)) return Ntruth;
  /***
  if (!CONSP(x)) Err("set-equal?: not a list",E_notlist, x);
  if (!CONSP(y)) Err("set-equal?: not a list",E_notlist, y);
  ***/
  n=_length(x);
  m=_length(y);
  if(n!=m) return a_f;
  return subset(x,y);
}

ptr s_subset(int narg,ptr env)
{
  checknarg(narg-2);
   return subset(*(stack-2),*(stack-1));
}

ptr s_superset(int narg,ptr env)
{
  checknarg(narg-2);
   return subset(*(stack-1),*(stack-2));
}

/****     equal? or set_equal   ***/
ptr s_set_equal(int narg,ptr env)
{
  checknarg(narg-2);
   return equal_element(*(stack-2),*(stack-1));
}

/****  forall-element 
   return #t if test succeeds for all element of a set 
   else return #f
*****/
ptr s_forall_element(int narg,ptr env)
{
  ptr set,test,x;
  checknarg(narg-2);
    set=*(stack-2);
    test=*(stack-1);
    psh(set);
    while(CONSP(set)){
      *(stack-1)=CAR(set);
      set=CDR(set);
      x=(ptr)funcall(test,1,a_nil);
      if(x==a_f){
         pull();
         return x;
      }
   }
  pull();
  return a_t;
}

/****  exists-element
   return the element if test succeeds for a elemnt of a set
   else return #f
*****/
ptr s_exists_element(int narg,ptr env)
{
  ptr set,test,x,y;
  checknarg(narg-2);
    set=*(stack-2);
    test=*(stack-1);
    psh(set);
    while(CONSP(set)){
      y=*(stack-1)=CAR(set);
      set=CDR(set);
      x=(ptr)funcall(test,1,a_nil);
      if(x!=a_f){
         pull();
         return y;  /*** Why can't *pull()  ????  ***/
      }
   }
  pull();
  return a_f;
}

