/* 
 * tkMain.c --
 *
 *	This file contains a generic main program for Tk-based applications.
 *	It can be used as-is for many applications, just by supplying a
 *	different appInitProc procedure for each specific application.
 *	Or, it can be used as a template for creating new main programs
 *	for Tk applications.
 *
 * Copyright (c) 1990-1994 The Regents of the University of California.
 * Copyright (c) 1994-1997 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: tkMain.c,v 1.15 2002/12/13 16:54:35 dgp Exp $
 */
/**** 
   tk_tkmain.c
   chanegd for ss,  but not in TK original souce
   2003.10.05   first created
****/
#ifndef __WIN32__
/***  for cygwin  lisp.h conflict some win header ***/
#include "lisp.h"
#endif

#include <ctype.h>
#include <stdio.h> 
#include <string.h>
/** #define BUFSIZ 1024  * from stdio.h  ***/

#ifdef __WIN32__
#include <windows.h>
#define WINH_LOAD
#include "tcl.h"
#include "tk.h"
#include "tkWinInt.h"
#include "tclWinInt.h"
#include "lisp.h"
#else
#include <tcl.h>
/** #include <generic/tclInt.h>  **/
#include <tk.h>
/** #include <generic/tkInt.h>  **/
#endif

/** #include <stdio.h> **/
#ifdef NO_STDLIB_H
#   include "../compat/stdlib.h"
#else
#   include <stdlib.h>
#endif

extern int regcmd;
extern int verbose_mode;
extern Tcl_Interp *myInterp;
extern ptr  user_package;

typedef struct ThreadSpecificData {
    Tcl_Interp *interp;         /* Interpreter for this thread. */
    Tcl_DString command;        /* Used to assemble lines of terminal input
				 * into Tcl commands. */
    Tcl_DString line;           /* Used to read the next line from the
				 * terminal input. */
    int tty;                    /* Non-zero means standard input is a 
				 * terminal-like device.  Zero means it's
				 * a file. */
} ThreadSpecificData;
static Tcl_ThreadDataKey dataKey;

/*
 * Declarations for various library procedures and variables (don't want
 * to include tkInt.h or tkPort.h here, because people might copy this
 * file out of the Tk source directory to make their own modified versions).
 * Note: don't declare "exit" here even though a declaration is really
 * needed, because it will conflict with a declaration elsewhere on
 * some systems.
 */

#if !defined(__WIN32__) && !defined(_WIN32)
#if !defined(MAC_TCL)
extern int		isatty _ANSI_ARGS_((int fd));
#else
#include <unistd.h>
#endif
extern char *		strrchr _ANSI_ARGS_((CONST char *string, int c));
#endif

/*
 * Forward declarations for procedures defined later in this file.
 */

/*** void		ssPrompt(); ***/
static void		__StdinProc _ANSI_ARGS_((ClientData clientData,
			    int mask));
/*
 *----------------------------------------------------------------------
 *
 * Tk_MainEx --
 *
 *	Main program for Wish and most other Tk-based applications.
 *
 * Results:
 *	None. This procedure never returns (it exits the process when
 *	it's done.
 *
 * Side effects:
 *	This procedure initializes the Tk world and then starts
 *	interpreting commands;  almost anything could happen, depending
 *	on the script being interpreted.
 *
 *----------------------------------------------------------------------
 */
void
Tk_Mainex(argc, argv, appInitProc, interp)
    int argc;				/* Number of arguments. */
    char **argv;			/* Array of argument strings. */
    Tcl_AppInitProc *appInitProc;	/* Application-specific initialization
					 * procedure to call after most
					 * initialization but before starting
					 * to execute commands. */
    Tcl_Interp *interp;
{
    char *args;
    char buf[TCL_INTEGER_SPACE];
    int code,argc_memo;
    ptr s;
    size_t length;
    Tcl_Channel inChannel, outChannel;
    Tcl_DString argString;
    ThreadSpecificData *tsdPtr;
#ifdef __WIN32__
    HANDLE handle;
#endif
  toploop:  
        /** printf("Tk_Mainex enter interp  %x\n",interp); ***/
    /*
     * Ensure that we are getting the matching version of Tcl.  This is
     * really only an issue when Tk is loaded dynamically.
     */
     /****  sato  this cause abort!!! why?
    if (Tcl_InitStubs(interp, TCL_VERSION, 1) == NULL) {
	abort();
    }
    ****/

    tsdPtr = (ThreadSpecificData *) 
	Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
    
    Tcl_FindExecutable(argv[0]); /***  (tcl "info nameofexecutable")ǳǧǤ  ***/
    tsdPtr->interp = interp;

#if (defined(__WIN32__) || defined(MAC_TCL))
    Tk_InitConsoleChannels(interp);
#endif
    
#ifdef TCL_MEM_DEBUG
    Tcl_InitMemory(interp);
#endif

    /*
     * Parse command-line arguments.  A leading "-file" argument is
     * ignored (a historical relic from the distant past).  If the
     * next argument doesn't start with a "-" then strip it off and
     * use it as the name of a script file to process.
     */
     argc_memo=argc;
     argc=1; /***** drop Tcl script load  sato ***/
    if (argc > 1) {
	length = strlen(argv[1]);
	if ((length >= 2) && (strncmp(argv[1], "-file", length) == 0)) {
	    argc--;
	    argv++;
	}
    }
    if ((ptr)(long)TclGetStartupScriptFileName() == NULL) {
	if ((argc > 1) && (argv[1][0] != '-')) {
	    TclSetStartupScriptFileName(argv[1]);
	    argc--;
	    argv++;
	}
    }
    /*
     * Make command-line arguments available in the Tcl variables "argc"
     * and "argv".
     */
    args = Tcl_Merge(argc-1, (CONST char **)argv+1);
    Tcl_ExternalToUtfDString(NULL, args, -1, &argString);
    Tcl_SetVar(interp, "argv", Tcl_DStringValue(&argString), TCL_GLOBAL_ONLY);
    Tcl_DStringFree(&argString);
    ckfree(args);
    sprintf(buf, "%d", argc-1);

    if ((ptr)(long)TclGetStartupScriptFileName() == NULL) {
	Tcl_ExternalToUtfDString(NULL, argv[0], -1, &argString);
    } else {
	TclSetStartupScriptFileName(Tcl_ExternalToUtfDString(NULL,
		(char *)(long)TclGetStartupScriptFileName(), -1, &argString));
    }
    Tcl_SetVar(interp, "argc", buf, TCL_GLOBAL_ONLY);
    Tcl_SetVar(interp, "argv0", Tcl_DStringValue(&argString), TCL_GLOBAL_ONLY);

    /*
     * For now, under Windows, we assume we are not running as a console mode
     * app, so we need to use the GUI console.  In order to enable this, we
     * always claim to be running on a tty.  This probably isn't the right
     * way to do it.
     */

#ifdef __WIN32__
    handle = GetStdHandle(STD_INPUT_HANDLE);
    /*** lwritestr(STDMSG,"__WIN32__ handle\n"); ***/
    if ((handle == INVALID_HANDLE_VALUE) || (handle == 0) 
	     || (GetFileType(handle) == FILE_TYPE_UNKNOWN)) {
	/*
	 * If it's a bad or closed handle, then it's been connected
	 * to a wish console window.
	 */

	tsdPtr->tty = 1;
    } else if (GetFileType(handle) == FILE_TYPE_CHAR) {
	/*
	 * A character file handle is a tty by definition.
	 */

	tsdPtr->tty = 1;
    } else {
	tsdPtr->tty = 0;
    }

#else
    tsdPtr->tty = isatty(0);
#endif
/*** printf("Tk  debug 3\n");
getchar();
***/
    Tcl_SetVar(interp, "tcl_interactive",
	    (((ptr)(long)TclGetStartupScriptFileName() == NULL) 
		    && tsdPtr->tty) ? "1" : "0", TCL_GLOBAL_ONLY);

    /*
     * Invoke application-specific initialization.
     */

    if ((*appInitProc)(interp) != TCL_OK) {
	TkpDisplayWarning(Tcl_GetStringResult(interp),
		"Application initialization failed");
    }

    /*
     * Invoke the script specified on the command line, if any.
     */

    if ((ptr)(long)TclGetStartupScriptFileName() != NULL) {
	Tcl_ResetResult(interp);
	code = Tcl_EvalFile(interp, (char *)(long)TclGetStartupScriptFileName());
	if (code != TCL_OK) {
	    /*
	     * The following statement guarantees that the errorInfo
	     * variable is set properly.
	     */

	    Tcl_AddErrorInfo(interp, "");
	    TkpDisplayWarning(Tcl_GetVar(interp, "errorInfo",
		    TCL_GLOBAL_ONLY), "Error in startup script");
	    Tcl_DeleteInterp(interp);
	    Tcl_Exit(1);
	}
	tsdPtr->tty = 0;
    } else {

	/*
	 * Evaluate the .rc file, if one has been specified.
	 */

	Tcl_SourceRCFile(interp);

	/*
	 * Establish a channel handler for stdin.
	 */
        /****   chneged for ss  by sato ***/
      /****
	inChannel = Tcl_GetStdChannel(TCL_STDIN);
	if (inChannel) {
	    Tcl_CreateChannelHandler(inChannel, TCL_READABLE, __StdinProc,
		    (ClientData) inChannel);
	}
      ****/
      /***
	if (tsdPtr->tty) {
	    ssPrompt();
	}
      ***/
    }
    Tcl_DStringFree(&argString);

/***
    outChannel = Tcl_GetStdChannel(TCL_STDOUT);
    if (outChannel) {
	Tcl_Flush(outChannel);
    }
    Tcl_DStringInit(&tsdPtr->command);
    Tcl_DStringInit(&tsdPtr->line);
    Tcl_ResetResult(interp);
****/

    /*
     * Loop infinitely, waiting for commands to execute.  When there
     * are no windows left, Tk_MainLoop returns and we exit.
     */
/****  end of TK/generic/tkMain.c ***/
/*****   code for ss   here  ****/
/***
printf("Tk  debug 1\n");
getchar();
***/
    /***  XSynchronize(Tk_Display(Tk_MainWindow(interp)), TRUE);  from STk  ***/
      code = TkpInit(interp); /***  we don't call ***/
   code=TCL_OK;
  if (code == TCL_OK){
        if (verbose_mode)
            lwritestr(STDMSG,"TK  initialized! OK!\n");
    /*** Tk_initialized = 1;    Ok, it's fully initialized */
  }
  else{
      printf("***** Tk  Failed!  code %d  Why? ******\n", code);
       /****  tk8.3 is OK but 8.4 returns code 1  !!!! ***/
       exit(999); 
  }
#ifdef HAVE_LIBTKOGL
  if(Tkogl_Init(interp)!=TCL_OK){
      lwritestr(STDMSG,"***** Tk  OGL Failed! ******\n");
  }
  else if(verbose_mode)
      lwritestr(STDMSG,"TK/OGL OK!\n");
#endif

if(regcmd==0){
          /** s=intern_c("tk-init",a_unbdd); **/
          s=intern_c_pack("tk-init",a_unbdd,tk_package);
           /** eval_one(CONS(s,a_nil)); **   eval   Caution  ***/
           /*** meval(CONS(s,a_nil),a_nil);  **/
         /***
         PRINT("*****    tk:tk-init     *****");
          prin1(s,STDOUT);
          prin1(ATOMVAL(s),STDOUT);
         newline(STDOUT);
         ****/
     /***
      PRINT("(tk-init)---START");
      newline(STDOUT);
      ***/
           vm_eval(CONS(s,a_nil),a_nil);  /** V2.0 ***/
   if(verbose_mode){
      PRINT("(tk-init)---END");
      newline(STDOUT);
   }
          /*** s_gread(0,a_nil);    for debug  **/
           regcmd=1;
        /***  s_bye();  ***/
        }
     argc=argc_memo;
     /** Tcl_Eval(myInterp, "wm withdraw \".\""); ***/
     /***   No good!!!
     Tcl_Eval(myInterp, "wm geometry  \".\" 60x90+10+10");
     Tcl_ResetResult(myInterp);
     ****/

    top_visible=0; /*** Now not visible  ***/

      initload("tkinit");
      /****
      PRINT("tkinit---END");
      newline(STDOUT);
       ****/
      if(istty) load_verbose_set(YES);
      /*** printf(" tkinit OK\n"); ***/
     ATOMVAL(a_package)=user_package;
     make_prompt();
     paraload(argc,argv);
    s=CMAKESTR("start");
    psh(s);
    s=probe_ext(s,0);
    if(STREAMP(s)){
      sys_close(s);
      s_load(1,a_nil);
    }
    pull();
    if(top_visible) update();

    lwritestr(STDMSG,LOGINMSG);
    /**** istty=tsdPtr->tty;   ***/ 
    /*** printf("istty=tsdPtr->tty=%d\n",tsdPtr->tty); ***/
    /***  printf("tty %d\n",istty); 
          printf("Enter top M loop\n"); 
    ***/
    /*** top_visible=0; * Now not visible ??? ***/
    /*** mctop();     M stack loop   ****/
    s_TOP();   /****  V2.0   mc.c  loop   ****/
   /*** We don't use
    Tk_MainLoop();
   ****/
    exit(0);   /**** Not reach or ? SEOF read  in top  ****/
    goto toploop;
    /*** Tcl_Exit(0);  ***/
}
/*
 *----------------------------------------------------------------------
 *
 * StdinProc --
 *
 *	This procedure is invoked by the event dispatcher whenever
 *	standard input becomes readable.  It grabs the next line of
 *	input characters, adds them to a command being assembled, and
 *	executes the command if it's complete.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Could be almost arbitrary, depending on the command that's
 *	typed.
 *
 *----------------------------------------------------------------------
 */

    /* ARGSUSED */
/***
#define BUF_SIZE  4096

unsigned char stdin_buff[BUF_SIZE+1];
static unsigned char *buffer = stdin_buff,*readptr=stdin_buff;
static int readbytes=0;
int sread_flag=0;

unsigned char test_buff[BUF_SIZE+1];
unsigned char *write_test=test_buff;
***/

/****
int ss_getchar(){
  char *p=test_buff;
  int c;
 loop:
  c=sub_getchar();
    if(c==-1) {
      Tcl_DoOneEvent(TCL_FILE_EVENTS);
      c=Tcl_InputBuffered(Tcl_GetStdChannel(TCL_STDIN));
      if(c>0) __StdinProc(0,0);
     goto loop;
    }
  return c;
}

int sub_getchar(){
  int w=0;
 loop:
   w++;
    if(w>100) return -1;    *** Normally, never occur *
  if(readptr<buffer){
     --readbytes;
     return *readptr++;
  }
 else{
   readptr=buffer=stdin_buff;
  readbytes=0;
  *buffer=0;
 }
 goto loop;
}
****/
         
/*
 *----------------------------------------------------------------------
 *
 * Prompt --
 *
 *	Issue a prompt on standard output, or invoke a script
 *	to issue the prompt.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	A prompt gets output, and a Tcl script may be evaluated
 *	in interp.
 *
 *----------------------------------------------------------------------
 */
/***
void ssPrompt()
{
  lfflush(STDOUT);
   princ(prompt,STDOUT);
  lfflush(STDOUT);
}   
***/     

