;;;  http-engine.ss
;;;   html->visible
;;;   2005.1.20  revised
;;;   use html-auto.ss
;;;
(in-package "http")
(require "http-base")

(define *debug* #f)

(define *tags*  (list ())) ;;; tag-list-stack
(define *tag-error*  #f)
(define *tag-len*  0)
(define *text-win*  #f)
(define *tag-menu*  ())
(define *tag-form*  ())


(define *htmode* #f)

(define (init-html-vars)
  (set!  *tags*  (list ())) ;;; tag-list-stack
  (set!  *tag-error*  #f)
  (set! *htmode* (make-HTMODE))
  (set!  *tag-len*  0)
  (set!  *tag-menu*  ())
  (set!  *tag-form*  ())
)

(define (pull-tag)
  (if (pair? (cdr *tags*) ) (pop *tags* ) (set! *tag-error* 'tag-stack))
)

(define (get-tags)
   (let* ((x (car *tags*)))
      (cond ((eq? x ()) #f)
            ((null? (cdr x)) (car x))
            (else x)
      )
))

(define (push-tag x)
   (push (cons x (car *tags*)) *tags*)
)

(define (push-tags . x)
   (push (append  x (car *tags*)) *tags*)
)

(define (get-text-win)
    (SWEB-html *active-web*)  ;;;  html-win
)

(define (set-text-win)
        (set! *text-win* (get-text-win));;;  html-win
  *text-win*
)


(define (tag-insert str)
  (let* ((x str)(w *text-win*))
   (when w
      (if (eq? x #\newline)
             (funcall w 'insert 'insert x)
         (if (pair? x)
                (begin
                )
                (begin
                   (set! tag (get-tags))
                    (if tag 
                       (begin
                         (if *debug* (format #t "tag ~s~%" tag))
                         (funcall w 'insert 'insert str tag)
                       )
                       (funcall w 'insert 'insert str)
                    )
                )
          )
     )
  )
))

(define (push-tag-url adr)
  (let* ((abs 0)(tag 0)(tagurl 0)(w *text-win*)(sweb *active-web*))
        (set! abs (if (cdr adr) 'href_abs 'href_relative))
        (set! tag (tk-string abs *href-num*))
        (set! tagurl  (tk-string tag "url"))
        (inc *href-num*)
        (funcall w 'tag 'configure  tag )
        (if (cdr adr)
              (push (car adr) *href-abs*)
              (push (car adr) *href-relative*)
        )
        (funcall w 'tag 'bind  tag "<Button-1>" 
                (lambda() (http-get *URL*))
        )
        (funcall w 'tag 'bind  tagurl   "<Enter>"
                    ;;;(lambda()  (set! *URL* (car adr)))
            (lambda() (set! *active-web* sweb) (set! *URL* (car adr))
                (symbol-value-set (SWEB-urlvar sweb) (car adr)))
        )
        
        (funcall w 'tag 'lower  tagurl )
    ;;;   (funcall w 'insert 'insert (car adr) (list abs tag  tagurl 'popup))
     (push-tags abs tag  tagurl 'popup)
))


(define (tag-insert-error  str)
  (let* ((w *text-win*))
     (funcall w 'insert 'insert str 'error)
))

(define (tag-newline)
   (let* ((w *text-win*) (tb 0))
     (set! *tag-len* 0)
     (when w
        (funcall w 'insert 'insert #\newline)
        (when (pair? (HTMODE-table-stack *htmode*))
           (funcall w 'mark 'set 'insert (list 'insert 'lineend))
        )
     )
))

(define (tag-insert-win win  w h)
  (let* ((pos (funcall *text-win* 'index 'insert))
         (tstack (HTMODE-table-stack *htmode*))
         (table 0) (txtwin 0) 
         (conf 0) (as 0)
        )
    (funcall *text-win* 'window 'create 'insert :window win)
    (when (pair? tstack)
#|
      (if (pair? wh)
         (begin
             (set! w (car wh)) (set! h (cadr wh))
         )
         (begin
            (set! conf (funcall win 'configure))
          (format #t "insert-win ~s ~s~%" win conf)
            (set! as   (assq :width conf))
            (if as (set! w (cadddr as)))
            (set! as   (assq :height conf))
            (if as  (set! h  (cadddr as)))
         )
      )
|#
      (inc w) (inc h)     ;;;;;;  Caution!!!
      (set! table (car tstack))
      (set! txtwin (make-TXTWIN))
      (set-TXTWIN txtwin :pos pos  :width w :height h
               :row (TABLE-table-row table) :col (TABLE-table-col table))
      (set-TABLE table :win-stack (cons (cons win txtwin)  (TABLE-win-stack table)))
    )
))

(define (set-txtwin win w h)
  (let* ((tstack (HTMODE-table-stack *htmode*))
         (s 0)
         (wins 0)
        )
    (when (pair? tstack)
       (set! s (car tstack))
       (set! wins  (TABLE-win-stack s))
       (set! s (assq win wins))
       (if s
          (set-TXTWIN (cdr s) :width w :height h)
          (format #t "error set-txtwin ~s ~%" win)
       )
    )
)) 

;;;;  Note TCL/TK admits lower letter for th first byte of Widget!!!
;;;;      So, we needs ".s"
(define (gen-widget-name s)
  (let* ((n 1) (b (tk-name *text-win* ".s"  s)))
    (set! s b)
    (while (winfo 'exists s)
      (set! s (tk-name b n))
      (inc n)
    )
   s
))


(define (tag-html html-list)
  (let* ((x 0) (tag 0) (op 0))
     (set-text-win)
     (clear-html #t)
     (init-html-vars)
     (set! html-list (html-analy html-list))
   (while (pair? html-list)
     (set! x (pop html-list))
     (if (and *debug* (pair? x)) (format #t "~s ~%" x tag))
     (if (pair? x)
       (begin
          (set! op x)
          (if (procedure? (car op)) 
                 (set! html-list (funcall (car op) (cdr op) html-list))
               (cond ( (symbol? (car op))
                   (tag-insert (tk-string "UnKnown "  (car op) (prop->string (cdr op))))   )
                   ((eq? (car op) :tag/)
                        (set! op (cadr op))
                        (if (procedure? op)  (funcall op )  )
                      )
                   ((eq? (car op) :comment)
                      (format #t "comment ~s~%" (cadr op))  );;; modify!!!
                   (else   
                      (format #t "???? ~s~%"  op) 
                       (if (string? (cdr op)) (tag-insert-error (cdr op)))
                     );;; modify!!!
               )
          ) 
       )
       (if (string? x)
                (when (not(space-only x))
                  (tag-insert x)  
                  (inc *tag-len*  (string-byte-length x))
                )
         ;;; newline () or  error
         (cond  ( (eq? x #\tab)         ;;; never occur
                   (tag-insert "   ")  
                   (inc *tag-len*  3)     )
                ( (eq? x #\newline)
                     (if (HTMODE-pre *htmode*) 
                        (tag-newline)
                        (tag-insert " ")  
                     )
                   )
                ((eq? x :error)
                   (tag-insert-error  "ERR") ;;; to be modify
                  )  
                ((eq? x :comment)  )
                ((eq? x ()) )   ;;; ignore ()
                (else
                     (format #t "HTML error: ~s~%" x) 
                    )  
         )
       )
   )
 )
))

(define (space-only str)
   (let*  ((x (string->list str)))
      (while (and (pair? x) (eq? (car x) #\space))
          (set! x (cdr x))
       )
   (eq? x ())
))

(define (prop->string p)
 (let* ((x 0) (dest 0) (res " "))
   (while (pair? p)
       (set! x (pop p))
       (set! dest (cdr x))
       (if dest
             (set! res   (tk-string res (car x)"="  dest " "))
             (set! res   (tk-string res (car x) " " ))
       )
    )
 res
))

(define (get-prop prop key default)
  (let* ((x (assq  key prop)))
     (if x 
         (cdr x)
          default
     )
))

;;;;
(define (t-html p htm)
   (tag-insert "html-begin ") htm  )

(define (t/html)
   (tag-insert "html-end")
 )

(define (t-body p htm)
   (tag-newline )
   (tag-insert "body: ")  htm )

(define (t-head p htm)
   (tag-newline )
   (tag-insert "head: ")  htm )

(define (t/head )
   (tag-newline )
   ;;;(tag-insert "/head")  
)

(define (t-base p htm)     ;;;;  must in head
   (let* ((x (assoc "href" p)))
     (if x (set! *Address-base* (cdr x)))
   htm
))


(define (t-title p htm)
   (tag-newline )
   (tag-insert "title: ") htm )

(define (t/title )
   (tag-newline )
)

(define (t-!doctype p htm)
   ;;;(format #t "doctype ~s~%" p)
   (tag-newline )
   (tag-insert "!doctype ") 
   ;;;(tag-insert p)   ;;;  we ignore all!
   (tag-newline )
htm )


(define (t-meta p htm)   ;;; charset=Shift_JIS
   (let* ((x 0)(s 0)(dest 0))
    (tag-insert  "META:  "   )
    (while (pair? p)
       (set! x (pop p))
       (set! dest (cdr x))
       (if dest
             (tag-insert  (tk-string (car x) dest))
             (tag-insert  (tk-string (car x) ))
       )
    )
   htm 
))

(define (t-address p htm)    ;;; renraku saki    italic
   (tag-newline)
  htm
) 

(define (t/address)
   (tag-newline)
)

(define (t-p p htm)   ;;; danraku
   (tag-newline)
   (tag-newline)
 htm )

(define (t-br p htm)   ;;; newline
   (tag-newline)
   htm )

(define (t-div p htm)   ;;; group
(tag-newline)
 htm )

(define (t/div)   
   (tag-newline)
 )

(define (t-span p htm)   ;;; group
htm
)

(define t-wbr t-br)

(define (t-nobr p htm)   ;;; newline prohibit
   htm )
;;;
(define (t-center p htm)   
   (push-tag 'center)
   htm )

(define (t/center )
   (pull-tag) 
   (tag-newline)
  )

;;;
#|
(define (t-hr p htm)   ;;; under line
   (push-tag 'underline)
   htm )

(define t/hr   pull-tag  )
|#

(define (t-sup p htm)   ;;; super script
   (push-tag 'super)
   htm )

(define t/sup   pull-tag  )


(define (t-sub p htm)   ;;; sub script
   (push-tag 'sub)
   htm )

(define t/sub   pull-tag  )


(define (t-em p htm)   ;;; emphasis  italic
   htm )


(define (t-strong p htm)   ;;; emphasis Bold  italic
   htm )

(define (t-dfn p htm)   ;;; emphasis Bold  
   htm )

(define (t-samp p htm)   ;;; sample   fixed 
   htm )

(define (t-code  p htm)   ;;; program code    fixed
   htm )


(define (t-kbd p htm)   ;;; keyboard   fixed
   htm )

(define (t-var p htm)   ;;; variable name    italic
   htm )

(define (t-cite p htm)   ;;; citation     italic
   htm )

(define (t-blink  p htm)   ;;; blink 
   (push-tags 'raised 'bg-snow)
   htm )

(define t/blink   pull-tag  )


(define (t-img p htm)   ;;; image source
   (let* ((x 0)(s 0)(dest 0))
    (set! x (assoc "alt" p))
    (if x
       (tag-insert  (cdr x) )
       (tag-insert  "IMG")
    )
  #|     
    (while (pair? p)
       (set! x (pop p))
       (set! dest (cdr x))
       (if dest
             (tag-insert  (tk-string (car x) dest))  
             (tag-insert  (tk-string (car x) ))  
       )
    )
  |#
   htm 
))

(define (t-h1 p htm)   ;;; midasi
   (tag-newline)
   (push-tag  'H1) 
   htm )

(define (t/h1 )   ;;; midasi
   (pull-tag)
   (tag-newline)
 )

(define (t-h2 p htm)   ;;; midasi
   (tag-newline)
   (push-tag  'H2) 
   htm )

(define (t-h3 p htm)   ;;; midasi
   (tag-newline)
   (push-tag  'H3) 
   htm )

(define (t-h4 p htm)   ;;; midasi
   (tag-newline)
   (push-tag  'H4) 
   htm )

(define (t-h5 p htm)   ;;; midasi
   (tag-newline)
   (push-tag  'H5) 
   htm )

(define (t-h6 p htm)   ;;; midasi
   (tag-newline)
   (push-tag  'H6) 
   htm )

(define t/h2 t/h1)
(define t/h3 t/h1)
(define t/h4 t/h1)
(define t/h5 t/h1)
(define t/h6 t/h1)

(define (t-b p htm)  ;;; bold
   (push-tag 'solid)
   htm
)

(define (t/b)
  (pull-tag)
)

(define (t-i p htm)  ;;; italic
   htm
)

(define (t-tt p htm)  ;;; type
   htm
)

(define (t-big p htm)  ;;; Big char
   htm
)

(define (t-small p htm)  ;;; small char
   htm
)

(define (t-blockquote p htm)  ;;; indent block
   htm
)

(define (t-ol p htm)  ;;; list
   (set-HTMODE *htmode* :list 'ol)
   htm
)

(define (t-li p htm)  ;;; list number for ol , maru for ul
  (let* ((m (HTMODE-list *htmode*)) (n 0))
  (tag-newline )
    (if (eq? m 'ol)
       (begin
          (set! n (HTMODE-linum *htmode*))
          (inc n)
          (set-HTMODE *htmode* :linum n)
         (tag-insert (tk-string n " ")) 
       )
      (tag-insert " ")
    )        
   htm
))

(define (t-ul p htm)  ;;; list 
   (set-HTMODE *htmode* :list 'ul)
   htm
)

(define (t/ul )  ;;; list
   (set-HTMODE *htmode* :list #f :linum 0)
)

(define t/ol t/ul)

(define (t-dl p htm)  ;;; definition list
  (tag-newline )
  htm
)

(define (t/dl )  ;;; definition list
  (tag-newline )
)


(define (t-dt  p htm)  ;;; definition word
  (tag-newline )
  htm
)

(define (t-dd  p htm)  ;;; definition description
  (tag-insert "   " )
  htm
)


(define (t-font p htm)  ;;; font  size="3"
  (let* ((x (assoc "size" p)))
     (when x
        (set! x (string->number (cdr x)))
        (when x
         (if (< x 0)
            (set! x (+ 3  x))
         )
         (if (or (< x 1) (> x 8)) (set! x #f))
        )
     )
    (if x (push-tag (vector-ref  *font-vector* x)) (push-tags))
   htm
))

(define (t/font)
   (pull-tag)
)

(define (t-basefont p htm)  ;;; font  size="3"
   htm
)

(define (t-hr p htm)  ;;; yoko keisen
  (tag-newline )
  (tag-insert "---------------------------------------------")
  (tag-newline )
   htm
)

(define (t-!-- p htm)  ;;; comment  ignore
   htm
)
;;;
(define (t-pre p  htm)    ;;; raw text  
    (set-HTMODE  *htmode* :pre  #t)
    htm
)

(define (t/pre )    ;;; 
    (set-HTMODE  *htmode* :pre  #f)
)



(define (t-a p htm)    ;;; s=name or href
  (let* ((x 0)(s 0) (dest 0)(url 0)(adr 0)(prop p) (out "href ") )
    (while (pair? prop)
       (set! x (pop prop))
       (set! s (car x))
       (set! dest (cdr x))
       (cond ( (eq?  s 'href) 
               (set! url (get-abs-url dest))
               (set! adr (cons url  (equal? url dest)))
               ;;;(tag-insert adr)
               (push-tag-url adr)
               (set-HTMODE *htmode* :href adr)
                  )
           #|
             (else
                 (tag-insert (tk-string out s "=" dest " "))
                )
           |#
       )
    )
  htm
))

(define (t-frame  p htm)    ;;; frame only show link!!!
  (let* ((name 0)(s 0) (dest 0)(url 0))
    (set! url (get-prop p 'src #f))
    (set! name (get-prop p 'name "none"))
    (when url
       (set! url (get-abs-url url))
       (tag-insert (tk-string "frame name " name " src= " url))
    )
  htm
))


(define (t/a)
    (pull-tag)
    (set-HTMODE *htmode* :href #f)
    ;;;  (tag-insert "/A>" ) ignore ???
)

(define (skip-htm htm)
  (let* ((x 0)(cont #t))
  (while (and cont (pair? htm))
        (set! x (car htm))
        (if (or (eq? x #\newline) (null? x))    ;;; skip newline, ()
          (set! htm (cdr htm))
          (set! cont #f)
        )
  )
 htm
))

(define (check-tag htm  tag)
  (let* ((x #\newline))
    (if (pair? htm)
     (begin
        (set! htm (skip-htm htm))
        (set! x (pop htm))
        (if  (and (pair? x)(eq? (car x) tag))
           (cons htm (cdr x))  ;;;; return next htm & property
           #f
        )
     )
    #f
   )
))

(define (check-tag-end htm tag)
  (let* ((x #\newline))
    (if (pair? htm)
     (begin
        (set! htm (skip-htm htm))
        (set! x (pop htm))
 ;;; (format #t "check-tag-end ~s ~s ~s ~%" x tag htm)
      ;; (if  (and (pair? x)(eq? (car x) #\/) (eq? (cdr x) tag))
       (if  (and (pair? x)(eq? (car x) :tag/ ) (eq? (cadr x) tag))
            htm
             #f
      )
    )
   #f
  )
))

(define (get-string-html  htm)
 (let* ((x 0)(memo htm)(res ())(cont #t))
    (while (and cont (pair? htm))
      (set! x (pop htm))
      (cond ( (string? x)
                 (push x res))
            ( (or (eq? x #\newline)(null? x)))
            (else
               (set! cont #f)
               (push x htm)
            )
       )
    )
    (set! res (apply string-append (reverse res)))
   (cons htm res)
))


(define (t-label p htm)    ;;;s for
  (let* ((x 0)(s 0) (dest 0) (prop p) (out "label "))
    (set!  x (assq 'for prop))
    (when x 
       (set! dest (cdr x))
       (set! out (tk-string out  " = " dest " "))
    )
    (tag-insert out )
  htm
))



(define (t/label)
   ;;; (tag-insert "/LABEL" )
)


(define (t-tbody p htm)    ;;; indicate group in table
 htm     ;;; we ignore
)
(define t-thead t-tbody)
(define t-tfoot t-tbody)

