;;;(in-package "ss")
#|
(define-syntax let (syntax-rules (begin)
   ((let () b ...)   (begin b ...))
   ((let ((x y)) b ...)  (let* ((x y)) b ...))
   ((let ((x y)(z u) ...) b ...)
      (let* ((temp y) )
         (let ((z u) ...)
           (let* ((x temp))  b ...))))
     ;;;   named let
   ((let name ((var init)...)
        body ...)
     (letrec ((name 
                (lambda (var ...) body ... )))
       (name init ...)
      )
    )   
  )
)
|#
(define-syntax or
   (syntax-rules ()
     ((or) #f)
     ((or x) x)
     ((or x y ...)
       (let* ((temp x))
         (if temp
              temp
              (or y ...)
         ))
     )
   ))

(define-syntax and
   (syntax-rules ()
     ((and) #t)
     ((and x) x)
     ((and x y ...)
         (if x
              (and y ...)
              #f
         )
     )
   ))

(define-syntax cond
  (syntax-rules (else)
    ((cond) #f)
    ((cond (else y ...)) (begin y ...))
    ((cond (test y ...)) (if test (begin y ...)))
    ((cond (test y ...) clause ...)
       (if test (begin y ...)
            (cond clause ...)
        )
      )
))

(define-syntax do
  (syntax-rules ()
    ((do ((var init step ...) ...)
         (test expr ...)
         command ...)
     (letrec
       ((loop
           (lambda (var ...)
             (if test
                 (begin
                   (if #f #f)
                   expr ...)
                 (begin command ...
                   (loop (do "step" var step ...)
                       ...)
                 )
              )
            )
           ))
        (loop init ...))  )
    ((do "step" x) x)
    ((do "step" x y) y)
   )
)

(define-syntax while
   (syntax-rules ()
    ( (while test expr ...)
      (letrec ((loop
                (lambda ()
                     (if test
                       (begin expr ... (loop))
                     )
                )
               ))
          (loop)
       )
    )
))

(define-syntax when
    (syntax-rules ()
      ((when test expr ...)
        (if test (begin expr ...)))
))

(define-syntax push
  (syntax-rules ()
    ((push x lis) (begin (set! lis (cons x lis)) lis))
))

(define-syntax pop
  (syntax-rules ()
    ((pop lis) (let* ((temp (car lis)))
                   (set! lis (cdr lis))
                   temp
                )
    )
))

(define-syntax inc
  (syntax-rules ()
    ((inc x ) (set! x (+ 1 x)))
    ((inc x n ) (set! x (+ n x)))
))

(define-syntax dec
  (syntax-rules ()
    ((dec x ) (set! x (- x 1 )))
    ((dec x n ) (set! x (- x n )))
))

(define-syntax foreach
  (syntax-rules ()
    ((foreach  var lis expr ... ) 
       (do ((var 0) (temp lis (cdr temp)))
           ((null? temp))
           (set! var (car temp))
           expr ...
       )
    )
))

(define-syntax case   ;;; R5RS p.40
  (syntax-rules (else)
   ((case (key ...) clauses ...)
      (let* ((atom-key (key ...)))
        (case atom-key clauses ...)))
  ((case key (else result1 result2 ...))
      (begin result1 result2 ...))
  ((case key ((atoms ...) result1 result2 ...))
    (if (memv key '(atoms ...))
        (begin result1 result2 ...)))
  ((case key ((atoms ...) result1 result2 ...)
                clause1  clause2 ...)  
      (if (memv key '(atoms ...))
        (begin result1 result2 ...)
        (case key clause1 clause2 ...)
      )
    )
  )
)

#|
(define-syntax call-with-current-continuation
  (syntax-rules ()
    ((call-with-current-continuation proc)  (call/cc proc))
  )
)
|#
#|  Why this wrong????
  ;;;;   not is a function when it compiled
  ;;;;   so, it goes wrong!!!
  ;;;;  Therefore TOP-level macro-bind needs in our system 
(define-syntax not (syntax-rules ()
  ((_ x)
    (if x #f #t)
   )
  )
)
|#
(define-syntax nth (syntax-rules ()
  ((nth lis pos)
    (list-ref  lis pos)
   )
  )
)

;;;;  catch-error
;;;; proc は引数なし。
;;;  error-procは２引数の関数である。
;;;;  帰り値は、procまたはerror-procの帰す値
;;; from lib.ss
(define-syntax catch-error
  (syntax-rules (*errorhook*)
   ((catch-error proc error-proc)
        (let* ((c (cons ()()))(ex 0)(memo *errorhook*))
          (set!  ex (call/cc (lambda(p) p)))
          (if (eq? ex c)
            (begin
              ;(format #t "error caught!~%")
              (set! c (error-proc (car c)(cdr c)))
              ;(format #t "error caught returns ~s~%" c)
            )
            (begin 
              (set! *errorhook* 
                  (lambda (enum p) (set-car! c enum)(set-cdr! c p) (ex c)))
              (set! c (proc))
            )
           )
          (set! *errorhook* memo) 
          c 
        ))
  )
)
;;;;  STk catch
(define-syntax catch
  (syntax-rules ()
    ((catch form ...)
        (catch-error (lambda () form ... #f) (lambda (e p) #t)))
  )
) 
;;; for tcl scan procedure
(define-syntax  scanf 
   (syntax-rules ()
      ( (scanf :do-rest inp)  ())
      ( (scanf :do-rest inp var1 var2 ...)  
             (begin (set! var1  (read inp)) (scanf :do-rest inp var2 ...)))
      ( (scanf  str var1 var2 ...)
         (let* ((in (open-input-string str)))
            (set! var1 (read in))
            (scanf  :do-rest in var2 ...)
            (close-input-port in)
         )
           )
   )
)
;;; (scanf "3.14 55 67" x y z)
(define-syntax  scan-list
   (syntax-rules ()
      ( (scan-list :do-rest inp)  ())
      ( (scan-list :do-rest inp var1 var2 ...)
             (begin (set! var1  (pop inp)) (scan-list :do-rest inp var2 ...)))
      ( (scan-list  lis var1 var2 ...)
         (let* ((in lis))
            (set! var1 (pop in))
            (scan-list  :do-rest in var2 ...)
         )
           )
   )
)
;;;;  (scan-list '(sato foo papa) x y z)

;;;;   Cf. GCL.ss
;;;;
;;;  (multiple-value-set!   v1 ...  form)
;;;;

(define-syntax multiple-value-set!
  (syntax-rules ()
     ((multiple-value-set! "false" )
          #f)
     ((multiple-value-set! "false" a vars ...)
        (begin
          (set! a #f)
          (multiple-value-set! "false" vars ...)
        )  )
     ((multiple-value-set! "rest" temp)
         #t
       )
     ((multiple-value-set! "rest" value x y ...)
        (if (null? value)
                 (multiple-value-set!  "false" x y ...)
               (begin
                 (set! x (car value))
                 (set! value (cdr value))
                 (multiple-value-set! "rest" value  y ...)
               )
           )
         )
     ((multiple-value-set! v x ...  proc)
        (let* ((temp (values->list proc)))
          (if (null? temp)
             (multiple-value-set!  "false" v x ...)
            (begin
              (set! v (car temp))
              (set! temp (cdr temp))
              (multiple-value-set! "rest" temp x ...)
            )
         )
        )
      )
   )
)


#| 
;;;OK
(macro-expand-rec ' (do ((var 0) (temp lis (cdr temp)))
           ((null? temp))
           (set! var (car temp))
           (write var)
           (write "why"))     ())

;;; this is OK!
(macro-expand-rec ' (do ((var 0) (temp lis (cdr temp)))
           ((null? temp))
           (set! var (car temp))
           (write "why"))     ())
;;;  OK
(macro-expand-rec ' (do ((var 0) (temp lis ))
           ((null? temp))
           (set! var (car temp))
           (write var)
           (write "why"))     ())

;;;  OK

(macro-expand-rec ' (do ((var 0) )
           ((null? temp))
           (set! var (car temp))
           (write var)
           (write "why"))     ())

(foreach x '(3 4 5) (write x))

(let ((temp '(4 5 6)))  (pop temp))

(case (* 2 3)
    ((2 3 5 7) 'prime)
    ((1 4 6 8 9) 'composite))
   ;;; composite
(case 'c
   ((a) 'a)
   ((b) 'b) )
   ;;; undef

(case (car '(c d))
   ((a e i o u) 'vowel)
   ((w y) 'semivowel)
   (else 'consonant) )
   ;;;consonant

(case 6
    ((2 3 5 7) 'prime)
    ((1 4 6 8 9) 'composite))
   ;;; composite

(let ((s 1)(i 0))
   (loop (if (= i 100) (return s)) (set! i (+ i 1)) (set! s (* s i)))
)
 (let* ((x 0)(y 0)(z 0))
       (multiple-value-set! x y z (values 3 5))
     (format #t "~s ~s ~s ~%" x y z)
 )

|#

