package require Tkogl

# Create OpenGL window
pack [OGLwin .gl -width 300 -height 300 ] -expand yes -fill both

# Initialization of viewing parameters
.gl eval \
	-matrixmode projection \
	-loadidentity \
	-ortho -5 5 -5 5 -5 5 \
	-matrixmode modelview \
	-loadidentity \
	-enable map1vertex3 \
	-mapgrid1 100 0 1 

# Initial control point coordinates
lappend cpoints {-4 -4 0} {-2 4 0} {2 -4 0} {4 4 0}

# Code to render curve and control points
set curve [.gl newlist]
set dots [.gl newlist]
proc computecurve {} {
    global curve cpoints dots
    catch {.gl deletelist $curve}
    catch {.gl deletelist dots}
    eval .gl newlist $curve -map1 map1vertex3 0 1 3 4 [join $cpoints]\
	    -evalmesh1 line 0 100
    for {set i 0} {$i < 4} {incr i} {
	eval lappend tmp -loadname $i \
		-begin points -vertex [lindex $cpoints $i] -end
    }
    eval .gl newlist $dots -initnames -pushname -1 -pointsize 4 $tmp
}
computecurve

# Code to figure out which control point was picked
proc pick_point { x y } {
    global cpoints dots pickdot
    set l [.gl select 100 -matrixmode projection \
	    -pushmatrix \
	    -loadidentity \
	    -pickmatrix $x $y 4 4\
	    -ortho -5 5 -5 5 -5 5 \
	    -matrixmode modelview \
	    -call $dots \
	    -matrixmode projection\
	    -popmatrix \
	    -matrixmode modelview ]
    set pickdot {}
    foreach hit $l {
	set pickdot [lindex $hit 3]
	if {$pickdot=={}} continue
	puts "vtx $pickdot"
	puts "world coords [.gl unproject $x $y 0]"
	break
    }
}

# Code to drag the picked control point
proc move_point { x y } {
    global cpoints pickdot
    if {$pickdot!={}} {
	set cpoints [lreplace $cpoints $pickdot $pickdot \
		[.gl unproject $x $y 0]]
	computecurve
	.gl redraw
    }
}

.gl main -clear colorbuffer \
	-color 1 1 1 \
	-call $curve \
	-call $dots 

bind .gl <Button-1> { pick_point %x %y }
bind .gl <B1-Motion> { move_point %x %y }
